package App::Yath::Plugin;
use strict;
use warnings;

our $VERSION = '0.001056'; # TRIAL

sub options {}

sub pre_init {}

sub post_init {}

sub post_run {}

sub find_files {}

sub block_default_search {}

sub claim_file {}

1;
